/*

	Quick and Dirty Game Development Framework (QDGDF)

	Copyright (C) 2001 Angel Ortega <angel@triptico.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	http://www.triptico.com

*/

#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <dsound.h>
#include <mmsystem.h>

#include "qdgdf_audio.h"


/******************
	Data
*******************/

LPDIRECTSOUND dsound;

/* * the sounds * */

#define MAX_SOUNDS 64

LPDIRECTSOUNDBUFFER _primary;
LPDIRECTSOUNDBUFFER _sounds[MAX_SOUNDS];

static int _sound_seq=0;

/* external hwnd */
HWND hwnd;

static int started=0;


/*******************
	Code
********************/

static int sound_test(void)
{
	DSBUFFERDESC dsbd;
	WAVEFORMATEX wf;

	if(!_qdgdfa_sound) return(0);

	if(!started)
	{
		/* deferred startup */

		/* gets hwnd from the outside by the
		   window title. What the hell these
		   Microsoft morons were thinking when they
		   associated a sound buffer to a WINDOW? */
		hwnd=FindWindow(NULL,_qdgdfa_window_title);

		if(hwnd==NULL)
		{
			printf("%s\n",_qdgdfa_window_title);
			_qdgdfa_sound=0;
			return(0);
		}

		if(dsound->lpVtbl->SetCooperativeLevel(dsound, hwnd,
			DSSCL_PRIORITY)!=DS_OK)
		{
			_qdgdfa_sound=0;
			return(0);
		}

		/* tries to set the primary buffer capabilities;
		   if it doesn't work, just use the defaults */
		memset(&dsbd, '\0', sizeof(dsbd));
		dsbd.dwSize=sizeof(dsbd);
		dsbd.dwFlags=DSBCAPS_PRIMARYBUFFER;
		dsbd.dwBufferBytes=0;
		dsbd.lpwfxFormat=NULL;

		memset(&wf, '\0', sizeof(wf));
		wf.wFormatTag=WAVE_FORMAT_PCM;
		wf.nChannels=1;
		wf.nSamplesPerSec=11025;
		wf.wBitsPerSample=16;
		wf.nBlockAlign=2;
		wf.nAvgBytesPerSec=wf.nSamplesPerSec*wf.nBlockAlign;

		if(dsound->lpVtbl->CreateSoundBuffer(dsound,
			&dsbd,&_primary,NULL)==DS_OK)
			_primary->lpVtbl->SetFormat(_primary,&wf);

		started=1;
	}

	/* everything is ok */
	return(1);
}


/* interface */

int qdgdfa_load_sound(char * wavfile)
{
	PCMWAVEFORMAT pcmwf;
	DSBUFFERDESC dsbd;
	FILE * f;
	int n,size,bits;
	char * w1;
	char * w2;
	unsigned long s1;
	unsigned long s2;

	if(!sound_test()) return(1);

	if((f=qdgdfa_load_wav(wavfile,&size,&bits))==NULL)
		return(-1);

	memset(&pcmwf,'\0',sizeof(pcmwf));

	pcmwf.wf.wFormatTag=WAVE_FORMAT_PCM;
	pcmwf.wf.nChannels=1;
	pcmwf.wf.nSamplesPerSec=11025;
	pcmwf.wf.nBlockAlign=bits/8;
	pcmwf.wf.nAvgBytesPerSec=
		pcmwf.wf.nSamplesPerSec*
		pcmwf.wf.nBlockAlign;
	pcmwf.wBitsPerSample=bits;

	memset(&dsbd,'\0',sizeof(dsbd));

	dsbd.dwSize=sizeof(DSBUFFERDESC);
	dsbd.dwFlags=DSBCAPS_CTRLVOLUME|
		     DSBCAPS_CTRLFREQUENCY|
		     DSBCAPS_CTRLPAN;
	dsbd.dwBufferBytes=size;
	dsbd.lpwfxFormat=(LPWAVEFORMATEX) &pcmwf;

	n=_sound_seq;

	if(dsound->lpVtbl->CreateSoundBuffer(dsound,&dsbd,
		&_sounds[n],NULL)!=DS_OK)
	{
		fclose(f);
		return(-2);
	}

	if(_sounds[n]->lpVtbl->Lock(_sounds[n],0,size,
		(void **)&w1,&s1,(void **)&w2,&s2,0)!=DS_OK)
	{
		fclose(f);
		return(-3);
	}

	if(w1!=NULL)
		fread(w1,1,s1,f);

	if(w2!=NULL)
		fread(w2,1,s2,f);

	fclose(f);

	if(_sounds[n]->lpVtbl->Unlock(_sounds[n],w1,s1,w2,s2)!=DS_OK)
		return(-4);

	_sound_seq++;
	return(n);
}


void qdgdfa_play_sound(int snd, int loop)
{
	if(!sound_test()) return;

	if(snd>=0 && snd<_sound_seq)
	{
		if(loop==0)
			_sounds[snd]->lpVtbl->Stop(_sounds[snd]);

		_sounds[snd]->lpVtbl->SetCurrentPosition(_sounds[snd],0);

		_sounds[snd]->lpVtbl->Play(_sounds[snd],0,0,
			loop ? DSBPLAY_LOOPING : 0);
	}
}


void qdgdfa_stop_sound(int snd)
{
	if(!sound_test()) return;

	if(snd>=0 && snd<_sound_seq)
		_sounds[snd]->lpVtbl->Stop(_sounds[snd]);
}


void qdgdfa_set_pan(int snd, int pan)
{
	if(!sound_test()) return;

	if(snd>=0 && snd<_sound_seq)
		_sounds[snd]->lpVtbl->SetPan(_sounds[snd],10000*pan);
}


void qdgdfa_reset(void)
{
	int n;

	if(!sound_test()) return;

	/* frees all sounds */
	for(n=0;n < MAX_SOUNDS;n++)
	{
		if(_sounds[n]!=NULL)
		{
			_sounds[n]->lpVtbl->Release(_sounds[n]);
			_sounds[n]=NULL;
		}
	}

	_sound_seq=0;
}


void qdgdfa_pause(int p)
{
	/* TO BE DONE */
}


void qdgdfa_startup(void)
{
	int n;

	if(!_qdgdfa_sound) return;

	for(n=0;n < MAX_SOUNDS;n++)
		_sounds[n]=NULL;

	if(DirectSoundCreate(NULL, &dsound, NULL)!=DS_OK)
	{
		_qdgdfa_sound=0;
		return;
	}

	atexit(qdgdfa_shutdown);
}


void qdgdfa_shutdown(void)
{
	if(!_qdgdfa_sound || !started)
		return;

	qdgdfa_reset();

	dsound->lpVtbl->Release(dsound);

	started=0;
}


void qdgdfa_restart(void)
{
	qdgdfa_shutdown();
	qdgdfa_startup();
}
