/*

	Quick and Dirty Game Development Framework (QDGDF)

	Copyright (C) 2001 Angel Ortega <angel@triptico.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	http://www.triptico.com

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "qdgdf_video.h"


/******************
	Data
******************/

/* use logger flag */
int _qdgdfv_use_logger=1;

/* logger file */
char _qdgdfv_logger_file[150]="";

/* the virtual screen */
unsigned char * _qdgdfv_virtual_screen=NULL;

/* clear color */
unsigned char _qdgdfv_clear_color=0;

/* screen sizes */
unsigned int _qdgdfv_screen_x_size=320;
unsigned int _qdgdfv_screen_y_size=200;

/* window position */
int _qdgdfv_window_x=60;
int _qdgdfv_window_y=60;

/* double mode (if available) */
int _qdgdfv_double_mode=0;

/* the window title (if available) */
char _qdgdfv_window_title[150]="qdgdf Window";

/* additional video mode info (private to driver) */
char _qdgdfv_additional_info[150]="";

/* additional video mode int info (private to driver) */
int _qdgdfv_additional_int_info=0;

/* the keys */

int _qdgdfv_key_up=0;
int _qdgdfv_key_down=0;
int _qdgdfv_key_left=0;
int _qdgdfv_key_right=0;
int _qdgdfv_key_escape=0;
int _qdgdfv_key_space=0;
int _qdgdfv_key_enter=0;
/* Dan Compton added these changes */
int _qdgdfv_key_control=0;
int _qdgdfv_key_shift_l = 0;
int _qdgdfv_key_shift_r = 0;
int _qdgdfv_key_alnum = 0;
char _qdgdfv_alnum = 0;
/* End changes */

/** static data **/

/* the palette. Includes by default the original palette from
   the never-finished game Freak Wars (256*rgb) */

unsigned int _qdgdfv_palette[256*3]={
    0, 0, 0, 28, 24, 12, 20, 16, 8, 68, 68, 68, 255, 224, 0, 24, 24, 24,
    16, 20, 20, 8, 12, 12, 4, 8, 8, 40, 52, 28, 32, 40, 16, 20, 28, 8,
    12, 24, 0, 68, 56, 40, 64, 48, 32, 56, 40, 24, 255, 69, 0, 220, 152, 152,
    216, 148, 148, 208, 136, 136, 204, 128, 128, 196, 120, 120, 196, 112, 112,
    188, 104, 104, 180, 96, 96, 176, 88, 88, 168, 84, 84, 164, 80, 80,
    160, 72, 72, 156, 64, 64, 148, 56, 56, 144, 56, 56, 136, 48, 48,
    132, 44, 44, 128, 40, 40, 124, 32, 32, 116, 28, 28, 112, 24, 24,
    104, 24, 24, 100, 20, 20, 96, 16, 16, 92, 12, 12, 84, 8, 8, 80, 8, 8,
    72, 8, 8, 68, 0, 0, 64, 0, 0, 255, 191, 0, 228, 212, 200, 228, 204, 188,
    228, 196, 180, 228, 188, 168, 228, 184, 160, 228, 180, 152, 228, 172, 140,
    228, 168, 132, 228, 160, 120, 220, 152, 112, 212, 148, 104, 204, 140, 96,
    196, 132, 88, 192, 124, 84, 184, 120, 76, 180, 116, 72, 168, 112, 68,
    160, 104, 64, 152, 100, 60, 144, 96, 56, 136, 88, 56, 128, 88, 52,
    120, 80, 48, 112, 76, 44, 104, 72, 40, 96, 64, 36, 84, 60, 32, 72, 56, 28,
    68, 52, 24, 56, 44, 24, 44, 40, 20, 36, 32, 16, 212, 216, 216,
    204, 208, 208, 196, 200, 200, 196, 196, 196, 188, 188, 188, 180, 184, 184,
    176, 180, 180, 168, 172, 172, 164, 164, 164, 160, 160, 160, 152, 152, 152,
    148, 152, 152, 140, 144, 144, 132, 136, 136, 132, 132, 132, 124, 124, 124,
    116, 120, 120, 112, 116, 116, 104, 108, 108, 100, 100, 100, 96, 96, 96,
    88, 88, 88, 80, 84, 84, 76, 80, 80, 68, 72, 72, 64, 64, 64, 60, 60, 60,
    52, 56, 56, 48, 52, 52, 40, 44, 44, 36, 36, 36, 32, 32, 32, 255, 89, 0,
    100, 216, 92, 92, 200, 88, 84, 184, 80, 80, 172, 72, 72, 156, 64,
    68, 144, 56, 60, 132, 52, 56, 120, 44, 48, 104, 40, 40, 88, 32,
    36, 76, 24, 28, 60, 24, 20, 48, 16, 16, 32, 12, 255, 29, 0, 168, 152, 128,
    164, 144, 120, 156, 136, 116, 148, 128, 108, 140, 120, 100, 136, 116, 96,
    132, 112, 88, 124, 104, 84, 116, 96, 80, 108, 88, 72, 104, 88, 68,
    100, 80, 60, 92, 76, 56, 84, 68, 52, 76, 60, 48, 72, 56, 44, 140, 120, 88,
    128, 108, 76, 116, 96, 68, 104, 88, 56, 92, 76, 48, 80, 64, 40,
    68, 56, 32, 60, 48, 24, 108, 116, 88, 100, 104, 80, 92, 96, 72,
    80, 88, 64, 72, 80, 56, 64, 72, 48, 56, 64, 40, 48, 56, 36, 228, 228, 104,
    208, 196, 80, 192, 168, 60, 172, 140, 44, 156, 112, 28, 136, 84, 20,
    120, 60, 8, 100, 40, 0, 228, 228, 228, 228, 196, 196, 228, 168, 168,
    228, 140, 140,  228, 112, 112,  228, 88, 88, 228, 56, 56, 228, 28, 28,
    228, 0, 0, 212, 0, 0, 200, 0, 0, 192, 0, 0, 180, 0, 0, 168, 0, 0,
    160, 0, 0, 148, 0, 0, 136, 0, 0, 124, 0, 0, 112, 0, 0, 100, 0, 0,
    92, 0, 0, 80, 0, 0, 68, 0, 0, 60, 0, 0, 204, 208, 228, 176, 180, 228,
    152, 152, 228, 128, 128, 228, 100, 104, 228, 72, 76, 228, 48, 52, 228,
    24, 24, 228, 0, 0, 228, 0, 0, 204, 0, 0, 184, 0, 0, 160, 0, 0, 140,
    0, 0, 120, 0, 0, 96, 0, 0, 76, 228, 228, 228, 228, 212, 196, 228, 192, 168,
    228, 180, 140, 228, 160, 112, 228, 148, 84, 228, 128, 56, 228, 116, 24,
    216, 104, 24, 208, 100, 16, 196, 92, 16, 192, 88, 12, 180, 80, 8,
    172, 72, 0, 164, 64, 0, 156, 60, 0, 228, 228, 228, 228, 228, 192,
    228, 228, 160, 228, 228, 128, 228, 228, 96, 228, 228, 64, 228, 228, 32,
    228, 228, 0, 148, 56, 0, 140, 52, 0, 132, 44, 0, 120, 32, 0, 68, 56, 36,
    60, 44, 24, 48, 32, 20, 40, 24, 12, 255, 204, 0, 0, 0, 64, 0, 0, 56,
    0, 0, 44, 0, 0, 32, 0, 0, 24, 0, 0, 12, 148, 96, 96, 228, 144, 60,
    228, 208, 68, 255, 181, 0, 255, 159, 0, 255, 139, 0, 255, 119, 0,
    255, 109, 0, 255, 0, 255 };



/******************
	Proto
******************/

/******************
	Code
*******************/


/** basic tools **/

char * qdgdfv_sprintf(char * fmt, ...)
/* static sprintf */
{
	static char sp_buf[4096];
	va_list argptr;

	va_start(argptr, fmt);
	vsprintf(sp_buf, fmt, argptr);
	va_end(argptr);

	return(sp_buf);
}


void qdgdfv_logger(char * where, char * msg)
/* logs a message */
{
	FILE * f;

	if(_qdgdfv_use_logger && _qdgdfv_logger_file[0]!='\0')
	{
		if((f=fopen(_qdgdfv_logger_file,"a"))!=NULL)
		{
			fprintf(f,"%s: %s\n",where,msg);

			fclose(f);
		}
	}
}


void volatile qdgdfv_bang(char * where, char * msg)
/* bangs on unrecoverable problems */
{
	qdgdfv_logger(where,msg);

	printf("\nBANG! %s: %s\n",where,msg);
	fflush(stdout);

	exit(1);
}


void * qdgdfv_malloc(int size)
/* 'secure' malloc */
{
	void * ptr;

	ptr=malloc(size);

	if(!ptr) qdgdfv_bang("a_malloc","Out of memory");

	return(ptr);
}


FILE * qdgdfv_fopen(char * file, char * mode)
/* 'secure' fopen */
{
	FILE * f;

	f=fopen(file,mode);

	if(f==NULL) qdgdfv_bang("qdgdfv_fopen: can't open",file);

	return(f);
}


/** resources **/

void qdgdfv_load_pcx(unsigned char * pcx, char * pcxfile, int size)
/* loads a pcx */
{
	int n,m;
	unsigned char c;
	FILE * f;

	f=qdgdfv_fopen(pcxfile,"rb");

	/* skips header */
	fseek(f,128,SEEK_SET);

	n=0;
	while(n<size)
	{
		c=fgetc(f);

		if(c>0xC0)
		{
			/* run-length */
			m=c & 0x3F;

			c=fgetc(f);
		}
		else
			m=1;

		while(m)
		{
			pcx[n++]=c;
			m--;
		}
	}

	fclose(f);
}


void qdgdfv_load_palette(char * palette_file)
/* loads the palette */
{
	int n;
	FILE * f;
	char lin[80];
	unsigned int r,g,b;

	if(*palette_file=='\0') return;

	f=qdgdfv_fopen(palette_file,"r");

	/* ignores first three lines */
	fgets(lin,sizeof(lin)-1,f);
	fgets(lin,sizeof(lin)-1,f);
	fgets(lin,sizeof(lin)-1,f);

	for(n=0;n < 256;n++)
	{
		if(fscanf(f,"%u %u %u\n",&r,&g,&b)!=3)
			break;

		_qdgdfv_palette[n*3]=r;
		_qdgdfv_palette[(n*3)+1]=g;
		_qdgdfv_palette[(n*3)+2]=b;
	}

	/* fills to the end with zeroes */
	for(;n < 256;n++)
	{
		_qdgdfv_palette[n*3]=0;
		_qdgdfv_palette[(n*3)+1]=0;
		_qdgdfv_palette[(n*3)+2]=0;
	}

	fclose(f);
}


void qdgdfv_clear_virtual_screen(void)
/* zeroes the virtual screen */
{
	memset(_qdgdfv_virtual_screen, _qdgdfv_clear_color,
		_qdgdfv_screen_x_size * _qdgdfv_screen_y_size);
}

