/*

	Quick and Dirty Game Development Framework (QDGDF)

	Copyright (C) 2001 Angel Ortega <angel@triptico.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	http://www.triptico.com

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>

#include <windows.h>
#include <ddraw.h>
#include <mmsystem.h>

#include "qdgdf_video.h"

#define NEWKEYCATCH 1

/******************
	Data
*******************/

HINSTANCE hinst;
HWND hwnd;
LPDIRECTDRAW ddraw;
LPDIRECTDRAWSURFACE dsurface;
LPDIRECTDRAWSURFACE dback;
DDSURFACEDESC ddsd;
DDSCAPS ddsc;
LPDIRECTDRAWPALETTE dpal;

#if !NEWKEYCATCH
/* keyboard scancodes */
static int _scan_codes[128];
#else
/* Dan Compton added these 3 variables */
static UINT vkey;
static int keyisdown;
static int oem_scan_code;
#endif
static unsigned char * _locked_screen;


/*******************
	Code
********************/

static void _double_dump_screen(void)
{
	int n,m;
	register char a;
	char * _scr;
	char * _v_scr;

	_scr=_locked_screen;
	_v_scr=_qdgdfv_virtual_screen;

	for(n=0;n < _qdgdfv_screen_y_size;n++)
	{
		for(m=0;m < _qdgdfv_screen_x_size;m++)
		{
			a=*_v_scr++;

			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
		}

		_scr+=(_qdgdfv_screen_x_size*2);
	}
}


/* non-static */

void qdgdfv_set_palette(void)
/* sets the palette */
{
	int n;
	PALETTEENTRY p[256];
	int * pal;

	pal=_qdgdfv_palette;

	for(n=0;n < 256;n++,pal+=3)
	{
		p[n].peRed=(*pal);
		p[n].peGreen=*(pal+1);
		p[n].peBlue=*(pal+2);
		p[n].peFlags=0;
	}

	if(ddraw->lpVtbl->CreatePalette(ddraw,DDPCAPS_8BIT|DDPCAPS_ALLOW256,
		p,&dpal,NULL)!=DD_OK)
		qdgdfv_bang("SetPalette","CreatePalette error");

	/* aplicar a la superficie */
	if(dsurface->lpVtbl->SetPalette(dsurface,dpal)!=DD_OK)
		qdgdfv_bang("SetPalette","SetPalette error");
}


void qdgdfv_dump_virtual_screen(void)
{
	/* locks */
	if(dsurface->lpVtbl->Lock(dsurface,NULL,&ddsd,
		DDLOCK_WAIT|DDLOCK_WRITEONLY,NULL)!=DD_OK)
	{
		/* restores */
		dsurface->lpVtbl->Restore(dsurface);

		/* tries again */
		if(dsurface->lpVtbl->Lock(dsurface,NULL,&ddsd,
			DDLOCK_WAIT|DDLOCK_WRITEONLY,NULL)!=DD_OK)
			qdgdfv_bang("DumpScreen","Lock error");
	}

	_locked_screen=(char *)ddsd.lpSurface;

	if(_qdgdfv_double_mode)
		_double_dump_screen();
	else
	{
		memcpy(_locked_screen,_qdgdfv_virtual_screen,
			_qdgdfv_screen_x_size * _qdgdfv_screen_y_size);
	}

	dsurface->lpVtbl->Unlock(dsurface,ddsd.lpSurface);
}


void qdgdfv_input_poll(void)
/* polls the input system */
{
	MSG msg;

	if(PeekMessage(&msg,NULL,0,0,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	#if NEWKEYCATCH
	switch(vkey) {
	case VK_UP: _qdgdfv_key_up = keyisdown; break;
	case VK_DOWN: _qdgdfv_key_down = keyisdown; break;
	case VK_LEFT: _qdgdfv_key_left = keyisdown; break;
	case VK_RIGHT: _qdgdfv_key_right = keyisdown; break;
	case VK_ESCAPE: _qdgdfv_key_right = keyisdown; break;
	case VK_SPACE: _qdgdfv_key_right = keyisdown; break;
	case VK_ENTER: _qdgdfv_key_right = keyisdown; break;
	case VK_CONTROL: _qdgdfv_key_control = keyisdown; break;
	case VK_SHIFT: 
		       if(oem_scan_code == 0x2a) _qdgdfv_key_shift_l = keyisdown;
		       else if(oem_scan_code == 0x36) _qdgdfv_key_shift_r = keyisdown;
		       break;
	default:
		       if(isalnum((unsigned char)vkey)) {
			       _qdgdfv_key_alnum = keyisdown;
			       _qdgdfv_alnum = vkey;
		       }
	}
	#else	    
	_qdgdfv_key_up=       _scan_codes[0x48];
	_qdgdfv_key_down=     _scan_codes[0x50];
	_qdgdfv_key_left=     _scan_codes[0x4b];
	_qdgdfv_key_right=    _scan_codes[0x3d];
	_qdgdfv_key_escape=   _scan_codes[0x01];
	_qdgdfv_key_space =   _scan_codes[0x39];
	_qdgdfv_key_enter =   _scan_codes[0x1c];
	_qdgdfv_key_control = _scan_codes[0x1d];
	/* _qdgdfv_key_control_r = _scan_codes[0x1d];
	   _qdgdfv_key_control_r=_scan_codes[0x36]; 
	   0x36 is right shift, windows does not differ between left and right control keys, both are 0x1d
	 */
	_qdgdfv_key_shift_l = _scan_codes[0x2a];
	_qdgdfv_key_shift_r = _scan_codes[0x36];
	#endif
			

}


#ifndef STDCALL
#define STDCALL __stdcall
#endif

long STDCALL WndProc(HWND hwnd, UINT msg, UINT wparam, LONG lparam)
{
	int k;

	switch(msg)
	{
	case WM_SETCURSOR:

		SetCursor(NULL);
		return(0);
	/* Dan Compton made changes to WM_KEYDOWN and WM_KEYUP message handlers */
	case WM_KEYDOWN:
		#if NEWKEYCATCH
		vkey = tolower(wparam);
		keyisdown = 1;
		oem_scan_code = (lparam >> 16) & 0xff;		
		#else
		k = (lparam >> 16) & 0xff;		
		_scan_codes[k] = 1;
		#endif
		return 0;
	case WM_KEYUP:
		#if NEWKEYCATCH
		vkey = tolower(wparam);
		keyisdown = 0;
		oem_scan_code = (lparam >> 16) & 0xff;
		#else		
		k = (lparam >> 16) & 0xff;
		_scan_codes[k] = 0;
		#endif
		return 0;
	case WM_CLOSE:
		DestroyWindow(hwnd);
		return(0);

	case WM_DESTROY:
		PostQuitMessage(0);
		return(0);

	case WM_SYSCOMMAND:

		switch(wparam)
		{
		case SC_SCREENSAVE:
		case SC_MONITORPOWER:

			/* disable screensavings */
			return(0);
		}

		break;

	case MM_MCINOTIFY:

/*		  vlogger("WndProc","MM_MCINOTIFY");
		_cd_playing=0; */
		break;
	}

	return(DefWindowProc(hwnd,msg,wparam,lparam));
}


void qdgdfv_startup(void)
{
	WNDCLASS wc;

	hinst=(HINSTANCE) _qdgdfv_additional_int_info;

	wc.style=CS_HREDRAW|CS_VREDRAW;
	wc.lpfnWndProc=WndProc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=hinst;
	wc.hIcon=NULL;
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=NULL;
	wc.lpszMenuName=NULL;
	wc.lpszClassName="qdgdf";

	if(!RegisterClass(&wc))
		qdgdfv_bang("VideoStartup","RegisterClass error");

	hwnd=CreateWindow("qdgdf",_qdgdfv_window_title,WS_POPUP,
		CW_USEDEFAULT,CW_USEDEFAULT,
		CW_USEDEFAULT,CW_USEDEFAULT,
		NULL,NULL,hinst,NULL);

	if(hwnd==NULL)
		qdgdfv_bang("VideoStartup","CreateWindow error");

	ShowWindow(hwnd,SW_SHOW);
	UpdateWindow(hwnd);
	SetCursor(NULL);

	if(DirectDrawCreate(NULL,&ddraw,NULL)!=DD_OK)
		qdgdfv_bang("VideoStartup","DirectDrawCreate error");

	if(ddraw->lpVtbl->SetCooperativeLevel(ddraw,hwnd,
		DDSCL_EXCLUSIVE|DDSCL_FULLSCREEN)!=DD_OK)
		qdgdfv_bang("VideoStartup","SetCooperativeLevel error");

	if(_qdgdfv_double_mode==0)
	{
		if(ddraw->lpVtbl->SetDisplayMode(ddraw,
			_qdgdfv_screen_x_size,
			_qdgdfv_screen_y_size,8)!=DD_OK)
		{
			_qdgdfv_double_mode=1;

			qdgdfv_logger("VideoStartup",
				"switching to double mode");
		}
	}

	/* this cannot be an else from previous if */

	if(_qdgdfv_double_mode==1)
	{
		if(ddraw->lpVtbl->SetDisplayMode(ddraw,
			_qdgdfv_screen_x_size*2,
			_qdgdfv_screen_y_size*2,8)!=DD_OK)
			qdgdfv_bang("VideoStartup","SetDisplayMode error");
	}

	/* primary surface */

	ddsd.dwSize=sizeof(ddsd);
	ddsd.dwFlags=DDSD_CAPS|DDSD_BACKBUFFERCOUNT;
	ddsd.ddsCaps.dwCaps=DDSCAPS_PRIMARYSURFACE|
			    DDSCAPS_FLIP|DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount=1;

	if(ddraw->lpVtbl->CreateSurface(ddraw,&ddsd,&dsurface,NULL)!=DD_OK)
		qdgdfv_bang("VideoStartup","CreateSurface error");

	/* obtains the backbuffer */

	ddsc.dwCaps=DDSCAPS_BACKBUFFER;

	if(dsurface->lpVtbl->GetAttachedSurface(dsurface,&ddsc,&dback)!=DD_OK)
		qdgdfv_bang("VideoStartup","GetAttachedSurface error");

	SetCursorPos(0,0);

	atexit(qdgdfv_shutdown);

	qdgdfv_set_palette();

	_qdgdfv_virtual_screen=(unsigned char *)
		qdgdfv_malloc(_qdgdfv_screen_x_size * _qdgdfv_screen_y_size);

	qdgdfv_clear_virtual_screen();
}


void qdgdfv_shutdown(void)
{
	static int _down=0;

	if(_down)
		return;

	dsurface->lpVtbl->Release(dsurface);
	ddraw->lpVtbl->Release(ddraw);

	SendMessage(hwnd,WM_CLOSE,0,0);

	_down=1;
}
