/*

	Quick and Dirty Game Development Framework (QDGDF)

	Copyright (C) 2001 Angel Ortega <angel@triptico.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	http://www.triptico.com

*/

#include <stdio.h>
#include <string.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifndef WITHOUT_MIT_EXTENSIONS
#include <X11/extensions/XShm.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#endif

#include "qdgdf_video.h"


/******************
	Data
*******************/

/* misc X11 info */

static Display * _x_display;
static int _x_screen;
static XVisualInfo _x_visualinfo;
static Window _x_window;
static Colormap _x_colormap;
static Visual * _x_visual;
static GC _x_gc;
static XEvent _x_event;
static XImage * _x_image;
static XColor _x_palette[256];
static int _x_bpp;
static int _x_pixelsize;

/* data for extended color depths */

static unsigned int _extended_palette[256];
static unsigned int _x_red_mask;
static unsigned int _x_green_mask;
static unsigned int _x_blue_mask;
static int _x_red_shift;
static int _x_green_shift;
static int _x_blue_shift;

/* the double screen */

static unsigned char * _double_screen;

static int _mit_extensions=0;

#ifndef WITHOUT_MIT_EXTENSIONS
int shm_major,shm_minor;
Bool shm_pixmaps;
XShmSegmentInfo _shm_seg_info;
#endif

/******************
	Code
*******************/

static void x_calc_shift(unsigned v, unsigned * mask, int * shift)
/* calcs masks and shifts */
{
	unsigned b;
	unsigned m;

	/* shift first */
	for(b=0x80000000,*shift=0;b;b>>=1,(*shift)++)
	{
		if(b & v)
			break;
	}

	/* now masks */
	for(m=0x80000000,*mask=0;b;b>>=1,(*mask)|=m,m>>=1)
	{
		if(!(b & v))
			break;
	}
}


/* many dumps in many sizes and color depths */

static void _double_dump_screen(char * screen)
{
	int n,m;
	register char a;
	char * _scr;
	char * _v_scr;

	_scr=_double_screen;
	_v_scr=screen;

	for(n=0;n<_qdgdfv_screen_y_size;n++)
	{
		for(m=0;m<_qdgdfv_screen_x_size;m++)
		{
			a=*_v_scr++;

			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
		}

		_scr+=(_qdgdfv_screen_x_size*2);
	}
}


static void _dump_pix2_screen(unsigned char * screen)
{
	unsigned short * dest;
	unsigned short c;
	int n;

	dest=(unsigned short *) _double_screen;

	for(n=0;n<_qdgdfv_screen_x_size*_qdgdfv_screen_y_size;n++)
	{
		c=(unsigned short) _extended_palette[*screen];
		*dest=c;

		screen++;
		dest++;
	}
}


static void _dump_pix4_screen(unsigned char * screen)
{
	unsigned int * dest;
	unsigned int c;
	int n;

	dest=(unsigned int *) _double_screen;

	for(n=0;n<_qdgdfv_screen_x_size*_qdgdfv_screen_y_size;n++)
	{
		c=_extended_palette[*screen];
		*dest=c;

		screen++;
		dest++;
	}
}


static void _double_dump_pix2_screen(unsigned char * screen)
{
	int n,m;
	unsigned short a;
	unsigned short * _scr;
	unsigned char * _v_scr;

	_scr=(unsigned short *) _double_screen;
	_v_scr=screen;

	for(n=0;n<_qdgdfv_screen_y_size;n++)
	{
		for(m=0;m<_qdgdfv_screen_x_size;m++)
		{
			a=(unsigned short) _extended_palette[*_v_scr++];

			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
		}

		_scr+=(_qdgdfv_screen_x_size*2);
	}
}


static void _double_dump_pix4_screen(unsigned char * screen)
{
	int n,m;
	unsigned int a;
	unsigned int * _scr;
	unsigned char * _v_scr;

	_scr=(unsigned int *) _double_screen;
	_v_scr=screen;

	for(n=0;n<_qdgdfv_screen_y_size;n++)
	{
		for(m=0;m<_qdgdfv_screen_x_size;m++)
		{
			a=_extended_palette[*_v_scr++];

			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
			*_scr=a;
			*(_scr+(_qdgdfv_screen_x_size*2))=a;
			_scr++;
		}

		_scr+=(_qdgdfv_screen_x_size*2);
	}
}


/* non-static */

void qdgdfv_set_palette(void)
/* sets the palette (in any of multiple video flavours) */
{
	int n;
	int * pal;

	pal=_qdgdfv_palette;

	if(_x_pixelsize==1)
	{
		for(n=0;n < 256;n++)
		{
			_x_palette[n].pixel=n;
			_x_palette[n].flags=DoRed|DoGreen|DoBlue;
			_x_palette[n].red=((short)pal[n*3])<<8;
			_x_palette[n].green=((short)pal[(n*3)+1])<<8;
			_x_palette[n].blue=((short)pal[(n*3)+2])<<8;
		}

		XStoreColors(_x_display, _x_colormap, _x_palette, 256);
	}
	else
	{
		unsigned int r,g,b;

		for(n=0;n < 256;n++)
		{
			r=((((unsigned)pal[n*3])<<24)&_x_red_mask)>>_x_red_shift;
			g=((((unsigned)pal[(n*3)+1])<<24)&_x_green_mask)>>_x_green_shift;
			b=((((unsigned)pal[(n*3)+2])<<24)&_x_blue_mask)>>_x_blue_shift;

			_extended_palette[n]=r|g|b;
		}
	}
}


/** interface **/

void qdgdfv_dump_virtual_screen(void)
{
	if(_qdgdfv_double_mode)
	{
		if(_x_pixelsize==1)
			_double_dump_screen(_qdgdfv_virtual_screen);
		else
		if(_x_pixelsize==2)
			_double_dump_pix2_screen(_qdgdfv_virtual_screen);
		else
		if(_x_pixelsize==4)
			_double_dump_pix4_screen(_qdgdfv_virtual_screen);

		if(_mit_extensions)
		{
#ifndef WITHOUT_MIT_EXTENSIONS
		XShmPutImage(_x_display, _x_window, _x_gc,
		_x_image, 0, 0, 0, 0, _qdgdfv_screen_x_size*2,
		_qdgdfv_screen_y_size*2, False);
#endif
		}
		else
		{
		XPutImage(_x_display, _x_window, _x_gc,
		_x_image, 0, 0, 0, 0, _qdgdfv_screen_x_size*2,
		_qdgdfv_screen_y_size*2);
		}
	}
	else
	{
		if(_x_pixelsize==1)
			_x_image->data=_qdgdfv_virtual_screen;
		else
		if(_x_pixelsize==2)
			_dump_pix2_screen(_qdgdfv_virtual_screen);
		else
		if(_x_pixelsize==4)
			_dump_pix4_screen(_qdgdfv_virtual_screen);

		if(_mit_extensions)
		{
#ifndef WITHOUT_MIT_EXTENSIONS
		XShmPutImage(_x_display, _x_window, _x_gc,
		_x_image, 0, 0, 0, 0, _qdgdfv_screen_x_size,
		_qdgdfv_screen_y_size, False);
#endif
		}
		else
		{
		XPutImage(_x_display, _x_window, _x_gc,
		_x_image, 0, 0, 0, 0, _qdgdfv_screen_x_size,
		_qdgdfv_screen_y_size);
		}
	}

	if(_mit_extensions)
		XFlush(_x_display);
	else
		XSync(_x_display, False);
}

/* macro doesnt require a function call, speed things up some in the input function */
#define qdgdfv_zero_2s_keys() { \
	if(_qdgdfv_key_up == 2) _qdgdfv_key_up = 0; \
	if(_qdgdfv_key_down == 2)  _qdgdfv_key_down = 0; \
	if(_qdgdfv_key_left == 2) _qdgdfv_key_left = 0; \
	if(_qdgdfv_key_right == 2) _qdgdfv_key_right = 0; \
	if(_qdgdfv_key_escape == 2) _qdgdfv_key_escape = 0; \
	if(_qdgdfv_key_space == 2) _qdgdfv_key_space = 0; \
	if(_qdgdfv_key_enter == 2) _qdgdfv_key_enter = 0; \
	if(_qdgdfv_key_control == 2) _qdgdfv_key_control = 0; \
	if(_qdgdfv_key_shift_l == 2) _qdgdfv_key_shift_l = 0; \
	if(_qdgdfv_key_shift_r == 2) _qdgdfv_key_shift_r = 0; \
	if(_qdgdfv_key_alnum == 2) _qdgdfv_key_alnum = 0; \
}

#define qdgdfv_key_set(keysym, s) { \
	switch((KeySym)keysym) \
	{ \
	case XK_Up: _qdgdfv_key_up=s; break; \
	case XK_Down: _qdgdfv_key_down=s; break; \
	case XK_Left: _qdgdfv_key_left=s; break; \
	case XK_Right: _qdgdfv_key_right=s; break; \
	case XK_Escape: _qdgdfv_key_escape=s; break; \
	case XK_space: _qdgdfv_key_space=s; break; \
	case XK_Return: _qdgdfv_key_enter=s; break; \
	/* Dan Compton added these changes */ \
	/* cannot distinguise between left and right control because windows doesn't */ \
	case XK_Control_L: \
	case XK_Control_R: _qdgdfv_key_control=s; break; \
	case XK_Shift_L: _qdgdfv_key_shift_l = s; break; \
	case XK_Shift_R: _qdgdfv_key_shift_r = s; break; \
	case XK_0: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '0'; break; \
	case XK_1: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '1'; break; \
	case XK_2: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '2'; break; \
	case XK_3: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '3'; break; \
	case XK_4: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '4'; break; \
	case XK_5: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '5'; break; \
	case XK_6: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '6'; break; \
	case XK_7: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '7'; break; \
	case XK_8: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '8'; break; \
	case XK_9: _qdgdfv_key_alnum = s; _qdgdfv_alnum = '9'; break; \
	case XK_a: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'a'; break; \
	case XK_b: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'b'; break; \
	case XK_c: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'c'; break; \
	case XK_d: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'd'; break; \
	case XK_e: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'e'; break; \
	case XK_f: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'f'; break; \
	case XK_g: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'g'; break; \
	case XK_h: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'h'; break; \
	case XK_i: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'i'; break; \
	case XK_j: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'j'; break; \
	case XK_k: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'k'; break; \
	case XK_l: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'l'; break; \
	case XK_m: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'm'; break; \
	case XK_n: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'n'; break; \
	case XK_o: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'o'; break; \
	case XK_p: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'p'; break; \
	case XK_q: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'q'; break; \
	case XK_r: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'r'; break; \
	case XK_s: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 's'; break; \
	case XK_t: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 't'; break; \
	case XK_u: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'u'; break; \
	case XK_v: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'v'; break; \
	case XK_w: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'w'; break; \
	case XK_x: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'x'; break; \
	case XK_y: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'y'; break; \
	case XK_z: _qdgdfv_key_alnum = s; _qdgdfv_alnum = 'z'; break; \
	} \
}

void qdgdfv_input_poll(void)
/* polls the input system */
{
	KeySym k;
	KeySym lastkeysym = 0;
	int state;

	qdgdfv_zero_2s_keys();

	while(XPending(_x_display))
	{
		XNextEvent(_x_display, &_x_event);

		if(_x_event.type == KeyPress) {
			lastkeysym = XKeycodeToKeysym(_x_display, _x_event.xkey.keycode, 0);
			qdgdfv_key_set(lastkeysym, 1);
		}
		else if(_x_event.type == KeyRelease) {
			k = XKeycodeToKeysym(_x_display, _x_event.xkey.keycode, 0);
			/* 2 indicates that the press will be removed next call */
			if(k == lastkeysym) state = 2;
			else state = 0;
			qdgdfv_key_set(k, state);
		}
	}
}

void qdgdfv_startup(void)
/* inits video system */
{
	unsigned long attrmask;
	XSetWindowAttributes attrs;
	XGCValues xgcvalues;
	XSizeHints hints;

	_x_display=XOpenDisplay(_qdgdfv_additional_info);

	if(!_x_display)
		qdgdfv_bang("qdgdfv_startup","Bad display");

#ifndef WITHOUT_MIT_EXTENSIONS
	if(XShmQueryVersion(_x_display,&shm_major,&shm_minor,&shm_pixmaps))
		_mit_extensions=1;
#endif

	_x_screen=DefaultScreen(_x_display);
	_x_bpp=DefaultDepth(_x_display,_x_screen);

	switch(_x_bpp)
	{
	case 8:  _x_pixelsize=1; break;
	case 15:
	case 16: _x_pixelsize=2; break;
	case 24:
	case 32: _x_pixelsize=4; break;
	}

	if(! XMatchVisualInfo(_x_display,_x_screen,_x_bpp,
		_x_bpp==8 ? PseudoColor : TrueColor ,&_x_visualinfo))
		qdgdfv_bang("qdgdfv_startup","XMatchVisualInfo");

	/* calculate color masks */
	if(_x_pixelsize>1)
	{
		x_calc_shift(_x_visualinfo.red_mask, &_x_red_mask, &_x_red_shift);
		x_calc_shift(_x_visualinfo.green_mask, &_x_green_mask, &_x_green_shift);
		x_calc_shift(_x_visualinfo.blue_mask, &_x_blue_mask, &_x_blue_shift);

		qdgdfv_logger("qdgdfv_startup",
			qdgdfv_sprintf("X11 TrueColor bpp: %d pixelsize: %d [%02X,%d,%02X,%d,%02X,%d]",
				_x_bpp, _x_pixelsize,
				_x_red_mask, _x_red_shift,
				_x_green_mask, _x_green_shift,
				_x_blue_mask, _x_blue_shift));
	}

	_x_visual=_x_visualinfo.visual;

	_x_colormap=XCreateColormap(_x_display,
		RootWindow(_x_display,_x_screen), _x_visual,
		_x_bpp==8 ? AllocAll : AllocNone );

	attrmask=CWEventMask|CWColormap|CWBorderPixel;
	attrs.event_mask=KeyPressMask|KeyReleaseMask|ExposureMask;
	attrs.colormap=_x_colormap;
	attrs.border_pixel=0;

	if(_qdgdfv_double_mode)
	{
		_x_window=XCreateWindow(_x_display,
		RootWindow(_x_display, _x_screen),
		_qdgdfv_window_x, _qdgdfv_window_y,
		_qdgdfv_screen_x_size*2, _qdgdfv_screen_y_size*2, 0, _x_bpp,
		InputOutput, _x_visual, attrmask, &attrs);

		hints.width=hints.min_width=hints.max_width=_qdgdfv_screen_x_size*2;
		hints.height=hints.min_height=hints.max_height=_qdgdfv_screen_y_size*2;
	}
	else
	{
		_x_window=XCreateWindow(_x_display,
		RootWindow(_x_display, _x_screen),
		_qdgdfv_window_x, _qdgdfv_window_y,
		_qdgdfv_screen_x_size, _qdgdfv_screen_y_size, 0, _x_bpp,
		InputOutput, _x_visual, attrmask, &attrs);

		hints.width=hints.min_width=hints.max_width=_qdgdfv_screen_x_size;
		hints.height=hints.min_height=hints.max_height=_qdgdfv_screen_y_size;
	}

	hints.flags=PPosition|PSize|PMinSize|PMaxSize;
	XSetWMNormalHints(_x_display, _x_window, &hints);

	xgcvalues.graphics_exposures=False;

	_x_gc=XCreateGC(_x_display, _x_window, GCGraphicsExposures,
		&xgcvalues);

	XMapWindow(_x_display, _x_window);

	for(;;)
	{
		XNextEvent(_x_display, &_x_event);

		if(_x_event.type==Expose && !_x_event.xexpose.count)
			break;
	}

	if(_qdgdfv_double_mode)
	{
		_double_screen=qdgdfv_malloc(_qdgdfv_screen_x_size*
			_qdgdfv_screen_y_size * 4 * _x_pixelsize);

		if(_mit_extensions)
		{
#ifndef WITHOUT_MIT_EXTENSIONS
		memset(&_shm_seg_info,'\0',sizeof(XShmSegmentInfo));

		_x_image=XShmCreateImage(_x_display, _x_visual,
		_x_bpp, ZPixmap, 0, &_shm_seg_info, 
		_qdgdfv_screen_x_size*2, _qdgdfv_screen_y_size*2);

		_shm_seg_info.shmid=shmget(IPC_PRIVATE,
			_x_image->bytes_per_line*_x_image->height,
			IPC_CREAT | 0777);

		if(_shm_seg_info.shmid<0) qdgdfv_bang("shmget","shmid<0");

		_shm_seg_info.shmaddr=shmat(_shm_seg_info.shmid,NULL,0);
		_shm_seg_info.readOnly=False;

		_x_image->data=_shm_seg_info.shmaddr;
#endif
		}
		else
		{
		_x_image=XCreateImage(_x_display, _x_visual,				
		_x_bpp, ZPixmap, 0, _double_screen,					
		_qdgdfv_screen_x_size*2, _qdgdfv_screen_y_size*2,			
		32, _qdgdfv_screen_x_size * 2 * _x_pixelsize);				
		}
	}
	else
	{
		if(_x_pixelsize!=1)
			_double_screen=qdgdfv_malloc(_qdgdfv_screen_x_size*
			_qdgdfv_screen_y_size * _x_pixelsize);

		if(_mit_extensions)
		{
#ifndef WITHOUT_MIT_EXTENSIONS
		memset(&_shm_seg_info,'\0',sizeof(XShmSegmentInfo));

		_x_image=XShmCreateImage(_x_display, _x_visual,
		_x_bpp, ZPixmap, 0, &_shm_seg_info, 
		_qdgdfv_screen_x_size, _qdgdfv_screen_y_size);

		_shm_seg_info.shmid=shmget(IPC_PRIVATE,
			_x_image->bytes_per_line*_x_image->height,
			IPC_CREAT | 0777);

		if(_shm_seg_info.shmid<0) qdgdfv_bang("shmget","shmid<0");

		_shm_seg_info.shmaddr=shmat(_shm_seg_info.shmid,NULL,0);
		_shm_seg_info.readOnly=False;

		_x_image->data=_shm_seg_info.shmaddr;
#endif
		}
		else
		{
		_x_image=XCreateImage(_x_display, _x_visual,
		_x_bpp, ZPixmap, 0,
		_x_pixelsize==1 ? _qdgdfv_virtual_screen : _double_screen,
		_qdgdfv_screen_x_size, _qdgdfv_screen_y_size, 32,
		_qdgdfv_screen_x_size * _x_pixelsize);
		}
	}

	XStoreName(_x_display, _x_window, _qdgdfv_window_title);

	if(_qdgdfv_window_x!=-1 && _qdgdfv_window_y!=-1)
		XMoveWindow(_x_display, _x_window,
		_qdgdfv_window_x, _qdgdfv_window_y);

	/* startup */

	qdgdfv_set_palette();

	if(_mit_extensions)
	{
#ifndef WITHOUT_MIT_EXTENSIONS
	if(_x_pixelsize==1 && !_qdgdfv_double_mode)
		_qdgdfv_virtual_screen=(unsigned char *) _x_image->data;
	else
	{
		_qdgdfv_virtual_screen=(unsigned char *)
		qdgdfv_malloc(_qdgdfv_screen_x_size * _qdgdfv_screen_y_size);
		_double_screen=_x_image->data;
	}

	XShmAttach(_x_display, &_shm_seg_info);

	qdgdfv_logger("qdgdfv_startup","using MIT X extensions");
#endif
	}
	else
	{
	_qdgdfv_virtual_screen=(unsigned char *)
		qdgdfv_malloc(_qdgdfv_screen_x_size * _qdgdfv_screen_y_size);
	}

	qdgdfv_clear_virtual_screen();
}


void qdgdfv_shutdown(void)
{
	if(_mit_extensions)
	{
#ifndef WITHOUT_MIT_EXTENSIONS
		XShmDetach(_x_display, &_shm_seg_info);
		XDestroyImage(_x_image);
		shmdt(_shm_seg_info.shmaddr);
		shmctl(_shm_seg_info.shmid, IPC_RMID, NULL);
#endif
	}
	else
		XDestroyImage(_x_image);

	XFreeGC(_x_display, _x_gc);
	XDestroyWindow(_x_display, _x_window);
	XCloseDisplay(_x_display);
}

