/*

	Quick and Dirty Pacman Game (QDPG) Version: 1.0

	Copyright (C) 2001 Daniel Compton  <C_Anon01@yahoo.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <ctype.h>
#include "intqueue.h"
#include "qdgdf_audio.h"
#include "qdgdf_video.h"
#include "qdgdf_video_ext.h"

#define LOOP_TIME 165000
/* 1,000,000,000 nanoseconds = 1,000,000 microseconds = 1,000 milliseconds = 1 second */

#define LEVEL_WIDTH 20
#define LEVEL_HEIGHT 15
#define LEVEL_SIZE (LEVEL_WIDTH * LEVEL_HEIGHT)

#define PACMAN_WIDTH 32
#define PACMAN_HEIGHT 32
#define PACMAN_SIZE (PACMAN_WIDTH * PACMAN_HEIGHT)

#define MONSTER_WIDTH PACMAN_WIDTH
#define MONSTER_HEIGHT PACMAN_HEIGHT
#define MONSTER_SIZE PACMAN_SIZE
#define POWERPAC_WIDTH PACMAN_WIDTH
#define POWERPAC_HEIGHT PACMAN_HEIGHT
#define POWERPAC_SIZE PACMAN_SIZE
#define PORTAL_WIDTH PACMAN_WIDTH
#define PORTAL_HEIGHT PACMAN_HEIGHT
#define PORTAL_SIZE PACMAN_SIZE
#define PPPACMAN_WIDTH PACMAN_WIDTH
#define PPPACMAN_HEIGHT PACMAN_HEIGHT
#define PPPACMAN_SIZE PACMAN_SIZE
#define PORTAL_WIDTH PACMAN_WIDTH
#define PORTAL_HEIGHT PACMAN_HEIGHT
#define PORTAL_SIZE PACMAN_SIZE

#define NUM_MONSTERS 4
#define NUM_PORTALS 10

#define NUM_PACMAN_PIX 6
#define NUM_PPPACMAN_PIX 6
#define NUM_MONSTER_PIX NUM_MONSTERS
#define NUM_PORTAL_PIX NUM_PORTALS

#define PACR 0
#define PACL 1
#define PACUR 2
#define PACUL 3
#define PACDR 4
#define PACDL 5

#define POWERPAC 0

#define PPPACR 0
#define PPPACL 1
#define PPPACUR 2
#define PPPACUL 3
#define PPPACDR 4
#define PPPACDL 5

#define MON1 0
#define MON2 1
#define MON3 2
#define MON4 3

#define GOLD_WIDTH (PACMAN_WIDTH / 3)
#define GOLD_HEIGHT (PACMAN_WIDTH / 3)

#define SCORE_GOLD 500

#define HWALL_WIDTH PACMAN_WIDTH
#define HWALL_HEIGHT (PACMAN_HEIGHT / 2)

#define VWALL_WIDTH (PACMAN_WIDTH / 2)
#define VWALL_HEIGHT PACMAN_HEIGHT

#define CWALL_WIDTHH (VWALL_WIDTH + VWALL_WIDTH / 2)
#define CWALL_HEIGHTH HWALL_HEIGHT
#define CWALL_WIDTHV VWALL_WIDTH
#define CWALL_HEIGHTV (HWALL_HEIGHT + HWALL_HEIGHT / 2)

static char *PacmanImages[NUM_PACMAN_PIX] = {
	"images/pacman_right32.pcx",
	"images/pacman_left32.pcx",
	"images/pacman_upr32.pcx",
	"images/pacman_upl32.pcx",
	"images/pacman_downr32.pcx",
	"images/pacman_downl32.pcx"
};

static char *PPPacmanImages[NUM_PPPACMAN_PIX] = {
	"images/pppacman_right32.pcx",
	"images/pppacman_left32.pcx",
	"images/pppacman_upr32.pcx",
	"images/pppacman_upl32.pcx",
	"images/pppacman_downr32.pcx",
	"images/pppacman_downl32.pcx"
};

static char *MonsterImages[NUM_MONSTER_PIX] = {
	"images/monster1.pcx",
	"images/monster2.pcx",
	"images/monster3.pcx",
	"images/monster4.pcx"
};

static char PowerPacImage[] = "images/powerpac.pcx";

static char *PortalImages[NUM_PORTAL_PIX] = { 
	"images/portal0_32.pcx", 
	"images/portal1_32.pcx", 
	"images/portal2_32.pcx", 
	"images/portal3_32.pcx", 
	"images/portal4_32.pcx", 
	"images/portal5_32.pcx", 
	"images/portal6_32.pcx", 
	"images/portal7_32.pcx", 
	"images/portal8_32.pcx", 
	"images/portal9_32.pcx"
};

/* In seconds */
#define MON1_INTERVAL 1
#define MON2_INTERVAL 1
#define MON3_INTERVAL 1
#define MON4_INTERVAL 1

#define PP_TIME 5

#define pacmanis(c) ((char)c == '@')
#define portalis(c) isdigit((unsigned char)c)
#define LevelXToScreen(x) ((unsigned int)x * PACMAN_WIDTH)
#define LevelYToScreen(y) ((unsigned int)y * PACMAN_HEIGHT)
#define GetBGChar(x, y) Level.map[(unsigned int)y][(unsigned int)x].bg
#define GetFGChar(x, y) Level.map[(unsigned int)y][(unsigned int)x].fg
#define GetCurrentBGChar() Level.map[Pacman.level_y][Pacman.level_x].bg
#define GetCurrentFGChar() Level.map[Pacman.level_y][Pacman.level_x].fg
#define GetCurrentMonsterBGChar(i) Level.map[Monster[(int)i].level_y][Monster[(int)i].level_x].bg
#define GetCurrentMonsterFGChar(i) Level.map[Monster[(int)i].level_y][Monster[(int)i].level_x].fg
#define GetPrevBGChar() Level.map[Pacman.prev_level_y][Pacman.prev_level_x].bg
#define GetPrevFGChar() Level.map[Pacman.prev_level_y][Pacman.prev_level_x].fg
#define GetPrevMonsterBGChar(i) Level.map[Monster[(int)i].prev_level_y][Monster[(int)i].prev_level_x].bg
#define GetPrevMonsterFGChar(i) Level.map[Monster[(int)i].prev_level_y][Monster[(int)i].prev_level_x].fg
#define EraseCurrentBGChar() (GetCurrentBGChar() = ' ')
#define EraseCurrentFGChar() (GetCurrentFGChar() = ' ')
#define ErasePrevBGChar() (GetPrevBGChar() = ' ')
#define ErasePrevFGChar() (GetPrevFGChar() = ' ')
#define ErasePrevMonsterBGChar(i) (GetPrevMonsterBGChar(i) = ' ')
#define ErasePrevMonsterFGChar(i) (GetPrevMonsterFGChar(i) = ' ')
#define SetCurrentBGChar(c) (GetCurrentBGChar() = (char)c)
#define SetCurrentFGChar(c) (GetCurrentFGChar() = (char)c)
#define SetCurrentMonsterFGChar(i, c) (GetCurrentMonsterFGChar((int)i) = (char)c)
#define SetCurrentMonsterBGChar(i, c) (GetCurrentMonsterBGChar((int)i) = (char)c)
#define PacmanCantBeHere() ( (wallis(GetCurrentBGChar())) || \
		( (Pacman.powerpac_set) && monsteris(GetCurrentFGChar()) ) )
#define MonsterCantBeHere(i) ( (wallis(GetCurrentMonsterBGChar((int)i))) || \
		(monsteris(GetCurrentMonsterFGChar((int)i))) || \
		(Pacman.powerpac_set && pacmanis(GetCurrentMonsterFGChar((int)i))))

#define RenderBothPacman() qdgdfv_display_pcx((!Pacman.powerpac_set)? \
					GameImages.pacman[GameImages.pacman_dir] : \
					GameImages.pppacman[GameImages.pacman_dir], \
					PACMAN_WIDTH, PACMAN_HEIGHT, \
					LevelXToScreen(Pacman.level_x), \
					LevelYToScreen(Pacman.level_y));

#define RenderPacman() qdgdfv_display_pcx(GameImages.pacman[GameImages.pacman_dir], \
					PACMAN_WIDTH, PACMAN_HEIGHT, \
					LevelXToScreen(Pacman.level_x), \
					LevelYToScreen(Pacman.level_y));

#define RenderPPPacman() qdgdfv_display_pcx(GameImages.pppacman[GameImages.pacman_dir], \
					PACMAN_WIDTH, PACMAN_HEIGHT, \
					LevelXToScreen(Pacman.level_x), \
					LevelYToScreen(Pacman.level_y));

#define RenderClearPrev() qdgdfv_clear_screen_at(LevelXToScreen(Pacman.prev_level_x), \
					LevelYToScreen(Pacman.prev_level_y), PACMAN_WIDTH, PACMAN_HEIGHT)

#define RenderMonster(i) qdgdfv_display_pcx(GameImages.monster[(int)i], \
					MONSTER_WIDTH, MONSTER_HEIGHT, \
					LevelXToScreen(Monster[(int)i].level_x), \
					LevelYToScreen(Monster[(int)i].level_y))

#define RIGHT 0
#define DOWN  1
#define LEFT  2
#define UP    3

/* Pacman */
struct Pacman_ {
	unsigned int level_x;
	unsigned int level_y;
	unsigned int prev_level_x;
	unsigned int prev_level_y;
	
	unsigned long score;
	int moved;

	int powerpac_set;
	int powerpac_used;
	time_t powerpac_time;
};
struct Pacman_ Pacman = {0};

struct Monster_ {
	int used; /* Boolean for whether this monster is in this map */
	unsigned int level_x;
	unsigned int level_y;
	unsigned int prev_level_x;
	unsigned int prev_level_y;
	int moved;
	double random_interval; /* Current interval before moving again */
	time_t time; /* Last time it moved */
	IntQueue *sight;
};
struct Monster_ Monster[NUM_MONSTERS] = {{0}};
IntQueue *Tempsight = NULL; /* For use by all the monsters - made global to reduce initializations */

struct GameImages_ {
	unsigned char pacman[NUM_PACMAN_PIX][PACMAN_SIZE]; /* static allocation because always load these */
	unsigned char *pppacman[NUM_PPPACMAN_PIX]; /* dynamic allocation cuz depends on map */
	unsigned char **pacman_display;
	int pacman_dir;

	unsigned char *monster[NUM_MONSTER_PIX]; /* dynamic allocation cuz depends on map */
	
	unsigned char *powerpac; /* dynamic allocation cuz depends on map */

	unsigned char *portal[NUM_PORTAL_PIX]; /* dynamic allocation cuz depends on map */
	
	unsigned char CLEAR_COLOR;
	unsigned char GOLD_COLOR;
	unsigned char HWALL_COLOR;
	unsigned char VWALL_COLOR;
	unsigned char CWALL_COLOR;
};
struct GameImages_ GameImages = {{{0}}};

typedef struct LevelTile_ {
	char bg; /* Background */
	char fg; /* Foreground */
} LevelTile;

struct Level_ {
	unsigned int gold_amount;
	int portals_used;
	LevelTile map[LEVEL_HEIGHT][LEVEL_WIDTH];
};
struct Level_ Level = {0};

struct Portal_ {
	char type; /* character which represents portal */
	unsigned int x1, y1; /* level coordinates */
	unsigned int x2, y2; /* level coordinates */
	int used; /* set if this portal is being used out of the NUM_PORTALS */
};
struct Portal_ Portal[NUM_PORTALS] = {{0}};

static int aborted = 0;

/* Function Prototypes */
void EndGame(char *message);

/* Functions */
#if 0
void CalibrateTime() {
	time_t starttime = time(NULL);
	time_t starttime2;
	/* the first loop effectively gets the next loop at teh start of the second */
	while(difftime((starttime2 = time(NULL)), starttime) < 1) { }
	while(difftime(time(NULL), starttime2) < 1) {
		tickssec++;
	}
	printf("Ticks/sec: %lu\n", tickssec);
}
#endif
int monsteris(char c) {
	switch(c) {
		case '#':
		case '$':
		case '%':
		case '^':
			return 1;
	}
	return 0;
}

int wallis(char c) {
	int result = 0;
	switch(c) {
		case '_':
		case '|':
		case '<':
		case '>':
		case '(':
		case ')':
			result = 1;
			break;
	}
	return result;
}

void CopyLineToMap(LevelTile *line, const char *buffer, size_t len) {
	size_t i = 0;
	for(i = 0; i < len; i++) {
		if((monsteris(buffer[i])) || (pacmanis(buffer[i])) ) {
			line[i].fg = buffer[i];
			line[i].bg = ' ';
		}
		else {
			line[i].fg = ' ';
			line[i].bg = buffer[i];
		}			
	}
}

int LoadLevel(char *path) {
	/* 1 for null and 1 for the \n */
	unsigned char buffer[LEVEL_WIDTH + 2];
	unsigned int line = 0;
	unsigned int i = 0;
	unsigned int numpacman = 0;
	FILE *f = fopen(path, "r");
	if(f == NULL) {
		puts("fopen failed - Possible level path invalid");
		return 0;
	}

	while(fgets(buffer, sizeof(buffer), f) != NULL) {
		char *nlptr;
		size_t slen;

		if(line >= LEVEL_HEIGHT) {
			puts("Level has too many lines");
			return 0;
		}

		nlptr = strchr(buffer, '\n');
		if( (nlptr != NULL) && ((slen = strlen(buffer)) == LEVEL_WIDTH + 1) ) *nlptr = 0;
		else {
			printf("Line #%d is too long or short\n", line + 1);
			return 0;
		}
		
		for(i = 0; i < LEVEL_WIDTH; i++) {
			switch(buffer[i]) {
				case ' ':
				case '_':
				case '|':
				case '<':
				case '>':
				case '(':
				case ')':
					break;
				case '#':
					Monster[MON1].used++;
					if(Monster[MON1].used > 1) {
						printf("Found extra monster 1(%c) on line %d, position %d\n",
						buffer[i], line + 1, i + 1);
						return 0;
					}
					Monster[MON1].level_x = i;
					Monster[MON1].level_y = line;
					Monster[MON1].prev_level_x = Monster[MON1].level_x;
					Monster[MON1].prev_level_y = Monster[MON1].level_y;
					break;
				case '$':
					Monster[MON2].used++;
					if(Monster[MON2].used > 1) {
						printf("Found extra monster 1(%c) on line %d, position %d\n",
						buffer[i], line + 1, i + 1);
						return 0;
					}
					Monster[MON2].level_x = i;
					Monster[MON2].level_y = line;
					Monster[MON2].prev_level_x = Monster[MON2].level_x;
					Monster[MON2].prev_level_y = Monster[MON2].level_y;
					break;
				case '%':
					Monster[MON3].used++;
					if(Monster[MON3].used > 1) {
						printf("Found extra monster 1(%c) on line %d, position %d\n",
						buffer[i], line + 1, i + 1);
						return 0;
					}
					Monster[MON3].level_x = i;
					Monster[MON3].level_y = line;
					Monster[MON3].prev_level_x = Monster[MON3].level_x;
					Monster[MON3].prev_level_y = Monster[MON3].level_y;
					break;
				case '^':
					Monster[MON4].used++;
					if(Monster[MON4].used > 1) {
						printf("Found extra monster 1(%c) on line %d, position %d\n",
						buffer[i], line + 1, i + 1);
						return 0;
					}
					Monster[MON4].level_x = i;
					Monster[MON4].level_y = line;
					Monster[MON4].prev_level_x = Monster[MON4].level_x;
					Monster[MON4].prev_level_y = Monster[MON4].level_y;
					break;
				case '!':
					Pacman.powerpac_used = 1;
					break;
				case '.':
					Level.gold_amount++;
					break;
				case '@':
					numpacman++;
					if(numpacman > 1) {
						printf(
						  "Found extra pacman(%c) on line %d, position %d\n",
						  buffer[i], line + 1, i + 1);
						return 0;
					}
					Pacman.level_x = i;
					Pacman.level_y = line;
					Pacman.prev_level_x = Pacman.level_x;
					Pacman.prev_level_y = Pacman.level_y;
					break;
				default:
					/* Obtain the portals */
					if(portalis(buffer[i])) {
						int val = buffer[i] - '0';
						Portal[val].used++;
						if(Portal[val].used == 1) {
							/* Dont think i will need the type */
							Portal[val].type = buffer[i];
							Portal[val].x1 = i;
							Portal[val].y1 = line;
						}
						else if(Portal[val].used == 2) {
							Portal[val].x2 = i;
							Portal[val].y2 = line;
						}
						else {
							printf("More then 2 instances of portal: %c\n",
									buffer[i]);
							return 0;
						}							
						/* May deem to be unnecessary */
						Level.portals_used++;
					}
					else {
						printf("Invalid character found in level: %c\n", buffer[i]);
						return 0;
					}
			}
		}

		CopyLineToMap(Level.map[line], buffer, LEVEL_WIDTH);
		line++;
	}	
	fclose(f);

	if(line != LEVEL_HEIGHT) {
		printf("Invalid length of level, only %d/%d lines\n", line, LEVEL_HEIGHT);
		return 0;
	}
	else if(numpacman == 0) {
		puts("Failure - No pacman found in level");
		return 0;
	}
	else if(Level.gold_amount == 0) {
		puts("Failure - No gold found in level");
		return 0;
	}

	for(i = 0; i < NUM_PORTALS; i++) {
		if(Portal[i].used == 1) {
			fprintf(stderr, "Portal %c missing a pair\n", Portal[i].type);
			return 0;
		}
	}
	
	return 1;
}

void InitColors() {
	GameImages.CLEAR_COLOR = qdgdfv_find_color(0x00, 0x00, 0x00);
	GameImages.GOLD_COLOR = qdgdfv_find_color(0xFF, 0x8C, 0x00);
	GameImages.HWALL_COLOR = qdgdfv_find_color(0x80, 0x00, 0x00);
	GameImages.VWALL_COLOR = GameImages.HWALL_COLOR;
	GameImages.CWALL_COLOR = GameImages.HWALL_COLOR;
}

void InitQdgdf() {
	/* Initialize the library values */
	_qdgdfa_16_bit = 1;
	_qdgdfv_screen_x_size = 640;
	_qdgdfv_screen_y_size = 480;
	_qdgdfv_clear_color = GameImages.CLEAR_COLOR;
	strcpy(_qdgdfv_window_title, "Quick & Dirty Pacman Game");
	strcpy(_qdgdfa_window_title, _qdgdfv_window_title);
	qdgdfa_startup();
	qdgdfv_startup();
}

void DestroyQdgdf() {
	qdgdfv_shutdown();
	qdgdfa_shutdown();
}

int LoadImages() {
	unsigned int i;

	/* Allocate space for the pcx images */
	for(i = 0; i < NUM_PACMAN_PIX; i++) {
		qdgdfv_load_pcx(GameImages.pacman[i], PacmanImages[i], PACMAN_SIZE);
	}

	if(Pacman.powerpac_used) {
		for(i = 0; i < NUM_PPPACMAN_PIX; i++) {
			GameImages.pppacman[i] = malloc(PPPACMAN_SIZE);
			if(GameImages.pppacman[i] == NULL) {
				fputs("malloc failure", stderr);
				return 0;
			}
			qdgdfv_load_pcx(GameImages.pppacman[i], PPPacmanImages[i], PPPACMAN_SIZE);
		}
		
		GameImages.powerpac = malloc(POWERPAC_SIZE);
		if(GameImages.powerpac == NULL) {
			fputs("malloc failure", stderr);
			return 0;
		}
		qdgdfv_load_pcx(GameImages.powerpac, PowerPacImage, POWERPAC_SIZE);
	}
	
	for(i = 0; i < NUM_MONSTER_PIX; i++) {
		if(Monster[i].used) {
			GameImages.monster[i] = malloc(MONSTER_SIZE);
			if(GameImages.monster[i] == NULL) {
				fputs("malloc failure", stderr);
				return 0;
			}
			qdgdfv_load_pcx(GameImages.monster[i], MonsterImages[i], MONSTER_SIZE);
		}
	}

	for(i = 0; i < NUM_PORTAL_PIX; i++) {
		if(Portal[i].used) {
			GameImages.portal[i] = malloc(PORTAL_SIZE);
			if(GameImages.portal[i] == NULL) {
				fputs("malloc failure", stderr);
				return 0;
			}
			qdgdfv_load_pcx(GameImages.portal[i], PortalImages[i], PORTAL_SIZE);
		}
	}
			


	/* Start the pacman looking to the right */
	GameImages.pacman_dir = PACR;
	return 1;
}

void FreeImages() {
	unsigned int i;

	if(Pacman.powerpac_used) {
		for(i = 0; i < NUM_PPPACMAN_PIX; i++) {
			free(GameImages.pppacman[i]);
		}
		free(GameImages.powerpac);
	}

	for(i = 0; i < NUM_MONSTER_PIX; i++) {
		if(Monster[i].used) {
			free(GameImages.monster[i]);
		}
	}

	for(i = 0; i < NUM_PORTAL_PIX; i++) {
		if(Portal[i].used) {
			free(GameImages.portal[i]);
		}
	}
}

int DisplayLevel() {
	unsigned int ly = 0;
	unsigned int lx;
	int i;
	time_t thetime;	

	qdgdfv_clear_virtual_screen();
	
	for(; ly < LEVEL_HEIGHT; ly++) {
		for(lx = 0; lx < LEVEL_WIDTH; lx++) {
			unsigned int x = LevelXToScreen(lx);
			unsigned int y = LevelYToScreen(ly);
			LevelTile c = Level.map[ly][lx];
			switch(c.bg) {
				case '.':
					qdgdfv_draw_rect_fill(x + PACMAN_WIDTH / 2 - GOLD_WIDTH / 2,
							y + PACMAN_HEIGHT / 2 - GOLD_HEIGHT / 2,
							GOLD_WIDTH, GOLD_HEIGHT, GameImages.GOLD_COLOR);
					break;
				case '_': /* wall will be an image later on */
					qdgdfv_draw_rect_fill(x, y + PACMAN_HEIGHT / 2 - HWALL_HEIGHT / 2,
							HWALL_WIDTH, HWALL_HEIGHT, GameImages.HWALL_COLOR);
					break;
				case '|': /* wall will be an image later on */
					qdgdfv_draw_rect_fill(x + PACMAN_WIDTH / 2 - VWALL_WIDTH / 2, y,
							VWALL_WIDTH, VWALL_HEIGHT, GameImages.VWALL_COLOR);
					break;
				case '<':
					qdgdfv_draw_rect_fill(x + (PACMAN_WIDTH - CWALL_WIDTHH),
							y + PACMAN_HEIGHT / 2 - CWALL_HEIGHTH / 2,
							CWALL_WIDTHH, CWALL_HEIGHTH, GameImages.CWALL_COLOR);
					qdgdfv_draw_rect_fill(x + PACMAN_WIDTH / 2 - CWALL_WIDTHV / 2,
							y + (PACMAN_HEIGHT - CWALL_HEIGHTV), CWALL_WIDTHV,
							CWALL_HEIGHTV, GameImages.CWALL_COLOR);
					break;
				case '>':
					qdgdfv_draw_rect_fill(x, y + PACMAN_HEIGHT / 2 - CWALL_HEIGHTH / 2,
							CWALL_WIDTHH, CWALL_HEIGHTH, GameImages.CWALL_COLOR);
					qdgdfv_draw_rect_fill(x + PACMAN_WIDTH / 2 - CWALL_WIDTHV / 2,
							y + (PACMAN_HEIGHT - CWALL_HEIGHTV), CWALL_WIDTHV,
							CWALL_HEIGHTV, GameImages.CWALL_COLOR);
					break;
				case '(':
					qdgdfv_draw_rect_fill(x + (PACMAN_WIDTH - CWALL_WIDTHH),
							y + PACMAN_HEIGHT / 2 - CWALL_HEIGHTH / 2,
							CWALL_WIDTHH, CWALL_HEIGHTH, GameImages.CWALL_COLOR);
					qdgdfv_draw_rect_fill(x + PACMAN_WIDTH / 2 - CWALL_WIDTHV / 2, y,
							CWALL_WIDTHV, CWALL_HEIGHTV, GameImages.CWALL_COLOR);
					break;	
				case ')':
					qdgdfv_draw_rect_fill(x, y + PACMAN_HEIGHT / 2 - CWALL_HEIGHTH / 2,
							CWALL_WIDTHH, CWALL_HEIGHTH, GameImages.CWALL_COLOR);
					qdgdfv_draw_rect_fill(x + PACMAN_WIDTH / 2 - CWALL_WIDTHV / 2, y,
							CWALL_WIDTHV, CWALL_HEIGHTV, GameImages.CWALL_COLOR);
					break;
				case '!':
					qdgdfv_display_pcx(GameImages.powerpac, POWERPAC_WIDTH,
							POWERPAC_HEIGHT, x, y);
					break;
				default:
					if(portalis(c.bg)) {
						int val = c.bg - '0'; 
						qdgdfv_display_pcx(GameImages.portal[val],
								PORTAL_WIDTH, PORTAL_HEIGHT,  x, y);
					
					}
			}
		}
	}	
	RenderPacman();

	thetime = time(NULL);
	
	for(i = MON1; i < NUM_MONSTERS; i++) {
		if(Monster[i].used) {
			qdgdfv_display_pcx(GameImages.monster[i], MONSTER_WIDTH, MONSTER_HEIGHT,
					LevelXToScreen(Monster[i].level_x),
					LevelYToScreen(Monster[i].level_y));
			Monster[i].time = thetime;
			Monster[i].sight = IQInit(15);
			if(Tempsight == NULL) {
				Tempsight = IQInit(15);
				if(Tempsight == NULL) {
					fputs("IntQueue initialization error", stderr);
					return 0;
				}
			}
			if(Monster[i].sight == NULL) {
				fputs("IntQueue initialization error", stderr);
				return 0;
			}
		}
	}

	qdgdfv_dump_virtual_screen();
	return 1;
}

void PacmanMoveRight() {
	Pacman.prev_level_x = Pacman.level_x;
	Pacman.prev_level_y = Pacman.level_y;
	if(Pacman.level_x < LEVEL_WIDTH - 1) {
		Pacman.level_x++;
	}
	else {
		Pacman.level_x = 0;
	}
}

void PacmanMoveLeft() {
	Pacman.prev_level_x = Pacman.level_x;
	Pacman.prev_level_y = Pacman.level_y;
	if(Pacman.level_x != 0) {
		Pacman.level_x--;
	}
	else {
		Pacman.level_x = LEVEL_WIDTH - 1;
	}
}

void PacmanMoveDown() {
	Pacman.prev_level_x = Pacman.level_x;
	Pacman.prev_level_y = Pacman.level_y;
	if(Pacman.level_y < LEVEL_HEIGHT - 1) {
		Pacman.level_y++;
	}
	else {
		Pacman.level_y = 0;
	}
}

void PacmanMoveUp() {
	Pacman.prev_level_x = Pacman.level_x;
	Pacman.prev_level_y = Pacman.level_y;
	if(Pacman.level_y != 0) {
		Pacman.level_y--;
	}
	else {
		Pacman.level_y = LEVEL_HEIGHT - 1;
	}
}

void PacmanMovePrev() {
	Pacman.level_x = Pacman.prev_level_x;
	Pacman.level_y = Pacman.prev_level_y;
}

void MoveToOtherPortal(char type) {
	int val = type - '0';
	Pacman.prev_level_x = Pacman.level_x;
	Pacman.prev_level_y = Pacman.level_y;
	if( (Pacman.level_x == Portal[val].x1) && (Pacman.level_y == Portal[val].y1) ) {
		Pacman.level_x = Portal[val].x2;
		Pacman.level_y = Portal[val].y2;
	}
	else {
		Pacman.level_x = Portal[val].x1;
		Pacman.level_y = Portal[val].y1;
	}
}
	
void MoveConsequences() {
	char bg = GetCurrentBGChar();
	char fg = GetCurrentFGChar();
	if(monsteris(fg)) {
		EndGame("The monster ate you, you lose!");
	}
	else {
		ErasePrevFGChar();
		switch(bg) {
			case '.':
				Level.gold_amount--;
				Pacman.score += SCORE_GOLD;
				EraseCurrentBGChar();
				if(Level.gold_amount == 0) { /* pacman wins */
					EndGame("You Win!");
				}				
				break;
			case '!': 
				Pacman.powerpac_set = 1;
				Pacman.powerpac_time = time(NULL);
				EraseCurrentBGChar();
				break;
			default:
				if(portalis(bg)) {
					char prev = GetPrevBGChar();
					/* update graphics to show move */
					if(portalis(prev)) {
						qdgdfv_display_pcx(GameImages.portal[prev - '0'],
							PORTAL_WIDTH, PORTAL_HEIGHT,
							LevelXToScreen(Pacman.prev_level_x), 
							LevelYToScreen(Pacman.prev_level_y));
					}
					else {
						RenderClearPrev();
					}
					RenderBothPacman();
					qdgdfv_dump_virtual_screen();
					/* move to other portal! */
					MoveToOtherPortal(bg);
				}
		}
		SetCurrentFGChar('@');
	}
	Pacman.moved = 1;
}
	
#define PacmanImageDirRight() (GameImages.pacman_dir = PACR)

#define  PacmanImageDirLeft() (GameImages.pacman_dir = PACL)

void PacmanImageDirDown() {
	switch(GameImages.pacman_dir) {
		case PACR:
			GameImages.pacman_dir = PACDL;
			break;
		case PACL:
			GameImages.pacman_dir = PACDR;
			break;
		case PACUR:
			GameImages.pacman_dir = PACDR;
			break;
		case PACUL:
			GameImages.pacman_dir = PACDL;
			break;
	}
}

void PacmanImageDirUp() {
	switch(GameImages.pacman_dir) {
		case PACR:
			GameImages.pacman_dir = PACUL;
			break;
		case PACL:
			GameImages.pacman_dir = PACUR;
			break;
		case PACDR:
			GameImages.pacman_dir = PACUR;
			break;
		case PACDL:
			GameImages.pacman_dir = PACUL;
			break;
	}
}

int ProcessInput() {
	int result = 1;

	Pacman.moved = 0;
	if(_qdgdfv_key_right) {
		PacmanMoveRight();
		PacmanImageDirRight();
		if(PacmanCantBeHere()) {
			PacmanMovePrev();
		}
		else {
			MoveConsequences();
		}
	}
	else if(_qdgdfv_key_left) {
		PacmanMoveLeft();
		PacmanImageDirLeft();
		if(PacmanCantBeHere()) {
			PacmanMovePrev();
		}
		else {
			MoveConsequences();
		}
	}		
	else if(_qdgdfv_key_up) {
		PacmanMoveUp();
		PacmanImageDirUp();
		if(PacmanCantBeHere()) {
			PacmanMovePrev();
		}
		else {
			MoveConsequences();
		}
	}
	else if(_qdgdfv_key_down) {
		PacmanMoveDown();
		PacmanImageDirDown();
		if(PacmanCantBeHere()) {
			PacmanMovePrev();
		}
		else {
			MoveConsequences();
		}
	}
	else if(_qdgdfv_key_escape) {
		EndGame("You Quit.");
		result = 0;
	}
	else result = 0;
	
	return result;
}

int LineSightPacman(unsigned int lx, unsigned int ly, int mon) {
	unsigned int tx;
	unsigned int ty;
	char tc;
	int right, down, left, up;
	/* the power pac makes the monsters run away */
	if(!Pacman.powerpac_set) {
		right = RIGHT;
		down = DOWN;
		left = LEFT;
		up = UP;
	}
	else {
		right = LEFT;
		down = UP;
		left = RIGHT;
		up = DOWN;
	}

	/* scan right */
	IQClear(Tempsight);
	for(tx = lx + 1, ty = ly; tx < LEVEL_WIDTH; tx++) {
		IQPush(Tempsight, right);
		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}
	
	/* scan right down (cant move right down so add down and right directions */
	IQClear(Tempsight);
	for(tx = lx + 1, ty = ly + 1; (tx < LEVEL_WIDTH) && (ty < LEVEL_HEIGHT); tx++, ty++) {
		if(wallis(GetBGChar(tx - 1, ty))) {
			IQPush(Tempsight, right);
			IQPush(Tempsight, down);
		}
		else {
			IQPush(Tempsight, down);
			IQPush(Tempsight, right);
		}

		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}
	
	/* scan down */
	IQClear(Tempsight);
	for(tx = lx, ty = ly + 1; ty < LEVEL_HEIGHT; ty++) {
		IQPush(Tempsight, down);
		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}

	/* scan left down */
	IQClear(Tempsight);
	tx = lx;
	ty = ly;
	while((tx != 0) && (ty < LEVEL_HEIGHT - 1)) {
		ty++;
		tc = GetBGChar(tx, ty);
		tx--;
		if(wallis(tc)) {
			IQPush(Tempsight, left);
			IQPush(Tempsight, down);
		}
		else {
			IQPush(Tempsight, down);
			IQPush(Tempsight, left);
		}

		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}
	
	/* scan left */
	IQClear(Tempsight);
	tx = lx;
	ty = ly;
	while(tx != 0) {
		tx--;
		IQPush(Tempsight, left);
		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}
	
	/* scan left up */
	IQClear(Tempsight);
	tx = lx;
	ty = ly;
	while((tx != 0) && (ty != 0)) {
		ty--;
		tc = GetBGChar(tx, ty);
		tx--;
		if(wallis(tc)) {
			IQPush(Tempsight, left);
			IQPush(Tempsight, up);
		}
		else {
			IQPush(Tempsight, up);
			IQPush(Tempsight, left);
		}

		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
		
	}
	
	/* scan up */
	IQClear(Tempsight);
	tx = lx;
	ty = ly;
	while(ty != 0) {
		ty--;
		IQPush(Tempsight, up);
		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}

	/* scan right up */
	IQClear(Tempsight);
	tx = lx;
	ty = ly;
	while((tx < LEVEL_WIDTH - 1) && (ty != 0)) {
		ty--;
		tc = GetBGChar(tx, ty);
		tx++;
		if(wallis(tc)) {
			IQPush(Tempsight, right);
			IQPush(Tempsight, up);
		}
		else {
			IQPush(Tempsight, up);
			IQPush(Tempsight, right);
		}

		if(wallis(GetBGChar(tx, ty))) {
			break;
		}
		else if(pacmanis(GetFGChar(tx, ty))) {
			IQCopy(Monster[mon].sight, Tempsight);
			return 1;
		}
	}

	/* if not found by this time, add a random direction to queue */
	IQPush(Monster[mon].sight, rand() % 4);
	

	return 0;
}

int NextDir(int dir) {
	dir--;
	if(dir < 0) dir = 3;
	return dir;
}

void MonsterMoveRight(int i) {
	Monster[i].prev_level_x = Monster[i].level_x;
	Monster[i].prev_level_y = Monster[i].level_y;
	if(Monster[i].level_x < LEVEL_WIDTH - 1) {
		Monster[i].level_x++;
	}
	else {
		Monster[i].level_x = 0;
	}
}

void MonsterMoveLeft(int i) {
	Monster[i].prev_level_x = Monster[i].level_x;
	Monster[i].prev_level_y = Monster[i].level_y;
	if(Monster[i].level_x != 0) {
		Monster[i].level_x--;
	}
	else {
		Monster[i].level_x = LEVEL_WIDTH - 1;
	}
}

void MonsterMoveDown(int i) {
	Monster[i].prev_level_x = Monster[i].level_x;
	Monster[i].prev_level_y = Monster[i].level_y;
	if(Monster[i].level_y < LEVEL_HEIGHT - 1) {
		Monster[i].level_y++;
	}
	else {
		Monster[i].level_y = 0;
	}
}

void MonsterMoveUp(int i) {
	Monster[i].prev_level_x = Monster[i].level_x;
	Monster[i].prev_level_y = Monster[i].level_y;
	if(Monster[i].level_y != 0) {
		Monster[i].level_y--;
	}
	else {
		Monster[i].level_y = LEVEL_HEIGHT - 1;
	}
}

void MonsterMovePrev(int i) {
	Monster[i].level_x = Monster[i].prev_level_x;
	Monster[i].level_y = Monster[i].prev_level_y;
}

void MonsterMoveConsequences(int i) {
	char fg = GetCurrentMonsterFGChar(i);
	ErasePrevMonsterFGChar(i);
	if(pacmanis(fg)) {
		EndGame("The monster ate you, you lose!");
	}
	switch(i) {
		case MON1: SetCurrentMonsterFGChar(i, '#'); break;
		case MON2: SetCurrentMonsterFGChar(i, '$'); break;
		case MON3: SetCurrentMonsterFGChar(i, '%'); break;
		case MON4: SetCurrentMonsterFGChar(i, '^'); break;
	}
	Monster[i].moved = 1;
}

int MoveMonsters() {
	time_t curtime = time(NULL);
	int dir;
	int result = 0;
	int i;
	int j;

	for(i = MON1; i < NUM_MONSTERS; i++) {
		if(Monster[i].used) {
			Monster[i].moved = 0;
			if(difftime(curtime, Monster[i].time) >= Monster[i].random_interval) {
				unsigned int interval = 1;
				char c;
				switch(i) {
					case MON1:
						interval = MON1_INTERVAL;
						break;
					case MON2:
						interval = MON2_INTERVAL;
						break;
					case MON3:
						interval = MON3_INTERVAL;
						break;
					case MON4:
						interval = MON4_INTERVAL;
						break;
				}
			
				/* perform AI to move this guy */
				LineSightPacman(Monster[i].level_x, Monster[i].level_y, i);
				IQPop(Monster[i].sight, &dir);
				j = 1;
				while( (!Monster[i].moved) && (j <= 4) ) {
					switch(dir) {
						case RIGHT:
							MonsterMoveRight(i);
							break;
						case DOWN:
							MonsterMoveDown(i);
							break;
						case LEFT:
							MonsterMoveLeft(i);
							break;
						case UP:
							MonsterMoveUp(i);
							break;
					}
					if(MonsterCantBeHere(i)) {
						MonsterMovePrev(i);
						dir = NextDir(dir);
					}
					else {
						MonsterMoveConsequences(i);
					}
					j++;
				}
				
	
				switch((c = GetPrevMonsterBGChar(i))) {
					case ' ':
						qdgdfv_clear_screen_at(
								LevelXToScreen(Monster[i].prev_level_x),
								LevelYToScreen(Monster[i].prev_level_y),
								MONSTER_WIDTH, MONSTER_HEIGHT);
						break;
					case '.':
						qdgdfv_clear_screen_at(
								LevelXToScreen(Monster[i].prev_level_x),
								LevelYToScreen(Monster[i].prev_level_y),
								MONSTER_WIDTH, MONSTER_HEIGHT);
						qdgdfv_draw_rect_fill(
							LevelXToScreen(Monster[i].prev_level_x)
								+ PACMAN_WIDTH / 2 - GOLD_WIDTH / 2,
							LevelYToScreen(Monster[i].prev_level_y)
								+ PACMAN_HEIGHT / 2 - GOLD_HEIGHT / 2,
							GOLD_WIDTH, GOLD_HEIGHT, GameImages.GOLD_COLOR);
						break;
					case '!':
						qdgdfv_display_pcx(GameImages.powerpac,
								POWERPAC_WIDTH, POWERPAC_HEIGHT,
								LevelXToScreen(Monster[i].prev_level_x),
								LevelYToScreen(Monster[i].prev_level_y));
						break;
					default:
						if(portalis(c)) {
							qdgdfv_display_pcx(GameImages.portal[c - '0'],
								PORTAL_WIDTH, PORTAL_HEIGHT,
								LevelXToScreen(Monster[i].prev_level_x),
								LevelYToScreen(Monster[i].prev_level_y));
						}
				}	
				RenderMonster(i);
			
				Monster[i].random_interval = (double)(rand() % interval);
				Monster[i].time = curtime;
				
				result = 1;
			}
			else if(Pacman.moved) {
				LineSightPacman(Monster[i].level_x, Monster[i].level_y, i);
			}
		}
	}
	
	return result;
}

void EndGame(char *message) {
	aborted = 1;
	puts(message);
}

void Cleanup() {
	int i;
	for(i = MON1; i < NUM_MONSTERS; i++) {
		if(Monster[i].used) {
			IQDestroy(Monster[i].sight);
			if(Tempsight != NULL) {
				IQDestroy(Tempsight);
				Tempsight = NULL;
			}
		}
	}
	FreeImages();
}


int main(int argc, char *argv[]) {
	if(argc == 2) {
		if(LoadLevel(argv[1]) == 0) {
			Cleanup();
			puts("The level specified is invalid");
			return EXIT_FAILURE;
		}
	}
	else {
		puts("No level specified - exiting");
		return EXIT_FAILURE;
	}

	srand((unsigned int)time(NULL));

	InitColors();
	
	InitQdgdf();
	
	if(LoadImages() == 0) {
		Cleanup();
		DestroyQdgdf();
		puts("Error occured in LoadImages");
		return EXIT_FAILURE;
	}

	if(DisplayLevel() == 0) {
		Cleanup();
		DestroyQdgdf();
		puts("Error occured in DisplayLevel");
		return EXIT_FAILURE;
	}
			
	while(!aborted) {
		struct timeval tv1;
		struct timeval tv2;
		struct timezone tz;
		struct timespec ts = {0};
		unsigned long usec;
		int procr;
		
		gettimeofday(&tv1, &tz);
		
		MoveMonsters();
		
		qdgdfv_input_poll();

		procr = ProcessInput();
		if(procr) {
			char prev = GetPrevBGChar();
			if(portalis(prev)) {
				qdgdfv_display_pcx(GameImages.portal[prev - '0'],
						PORTAL_WIDTH, PORTAL_HEIGHT,
						LevelXToScreen(Pacman.prev_level_x), 
						LevelYToScreen(Pacman.prev_level_y));
			}
			else {
				RenderClearPrev();
			}
			RenderBothPacman();
		}

		if( (Pacman.powerpac_set) && (difftime(time(NULL), Pacman.powerpac_time) >= PP_TIME) ) {
			Pacman.powerpac_set = 0;
			RenderPacman();
		}
		
			
		qdgdfv_dump_virtual_screen();

		gettimeofday(&tv2, &tz);
		usec = (tv2.tv_sec * 1000000 + tv2.tv_usec) - (tv1.tv_sec * 1000000 + tv1.tv_usec);
		ts.tv_nsec = usec < LOOP_TIME? ((LOOP_TIME - usec) * 1000) : 0;
		nanosleep(&ts, NULL);
	}

	printf("Your score is %lu\n", Pacman.score);

	Cleanup();
	DestroyQdgdf();
	return 0;
}

